#include <match_ip.h>

proc_print_out		print_out;
proc_print_raw		print_raw;
proc_matcher_add	add_matcher;

#define CHECK_IP(x) \
	(( x->layer[LAYER_NET].id == 0x0800 ) ? 1 : 0)

int ipid_match(void *priv, p_packet p, int negate)
{
	int ipid;
	p_range_priv x=(p_range_priv)priv;
	
	/* Make sure it's IP */
	if (!CHECK_IP(p)) return 0;

	/* Clamp values to maximum */
	if ( x->min > 65535 ) x->min=65535;
	if ( x->max > 65535 ) x->max=65535;
	
	ipid = ntohs(p->layer[LAYER_NET].h.ip->id);
	
	if ( (ipid >= x->min) && (ipid <= x->max) ) {
		MATCH_RET_POSITIVE;
	}else{
		MATCH_RET_NEGATIVE;
	}
}

int ttl_match(void *priv, p_packet p, int negate)
{
	int ttl;
	p_range_priv x=(p_range_priv)priv;
	
	/* Make sure it's IP */
	if (!CHECK_IP(p)) return 0;

	ttl=p->layer[LAYER_NET].h.ip->ttl;

	/* Clamp values to maximum */
	if ( x->min > 255 ) x->min=255;
	if ( x->max > 255 ) x->max=255;
	
	if ( (ttl >= x->min) && (ttl <= x->max) ) {
		MATCH_RET_POSITIVE;
	}else{
		MATCH_RET_NEGATIVE;
	}
}
uint _PLUGIN_INIT (p_plugin_in in, p_plugin_out out)
{
	PLUGIN_CHECK_PARAMS(in);

	print_out=in->print_out;
	print_raw=in->print_raw;
	add_matcher=in->add_matcher;

	out->name="match/ip";
	out->desc="Some generic IP based matchers";
	out->ver_major=1;
	out->ver_minor=0;
	
	add_matcher("TTL",
		in->templates.t_validate_range,
		in->templates.t_cleanup_ptr,
		ttl_match);
	
	add_matcher("IP_ID",
		in->templates.t_validate_range,
		in->templates.t_cleanup_ptr,
		ipid_match);
	
	return PLUGIN_ERR_OK;
}

/* This is where we would cleanup our shit */
uint _PLUGIN_UNLOAD (void)
{
	return PLUGIN_ERR_OK;
}
