#include <capdev.h>
#include <cleanup.h>
#include <s_globals.h>

/* Lock protects both lists */
p_captype captype=NULL;
p_capdev capdevs=NULL;
mlock_t captlock;


int capdev_init (void)
{
	MUTEX_CREATE(captlock, "Captlock");
	cleanup_add_handler(cleanup_lock, (void *)&captlock);
	return 1;
}


int capdev_add(char *type, char *interface, char *filter, proc_thread proc)
{
	p_captype t=NULL;
	p_capdev  d=malloc(sizeof(capdev_t));
	
	if ( !d ) return 0;

	if ( (t=capdev_find_type(type)) == NULL ) {
		return 0;
	}

	d->capture=t;
	d->state=CAP_STATE_NULL;
	d->interface=strdup(interface);
	d->filter=filter;
	d->priv=NULL; /* not set yet */
	d->thread_proc=proc;
	
	CRITICAL_START(captlock);
	capdevs=list_append(d, capdevs);
	CRITICAL_STOP(captlock);
	
	return 1;
}

void capdev_init_all(void)
{
	p_capdev dev;

	for (dev=capdevs; dev; dev=(p_capdev)dev->l.next) {
		if ( dev->state  != CAP_STATE_NULL ) continue;
		dev->priv = dev->capture->init(dev->interface, dev->filter);

		if ( dev->priv ) {
			dev->state = CAP_STATE_INIT;
		}else{
			dev->state = CAP_STATE_NULL;
		}
	}	
}

void capdev_start_all(void)
{
	p_capdev dev;

	for (dev=capdevs; dev; dev=(p_capdev)dev->l.next) {
		if ( dev->state != CAP_STATE_INIT ) continue;
		pthread_create (&dev->thread,NULL,dev->thread_proc,dev);

		if ( dev->priv ) {
			print_out("%s capture engine: started [%s]" CRLF,
				dev->capture->name, dev->interface);
			atomic_inc_counter(&globals.keepalive);
		}else{
			print_out("%s capture engine: failed [%s]" CRLF,
				dev->capture->name, dev->interface);
		}
	}

	cleanup_add_handler(capdev_end_all, NULL);
}

void capdev_end_all(int code, void *priv)
{
	p_captype tmp=NULL;
	p_captype foo=NULL;

	p_capdev dev=NULL;
	p_capdev bar=NULL;

	print_out("Stopping all capture devices" CRLF);
	
	CRITICAL_START(captlock);
	
	/* Remove all captypes */
	for (tmp=captype; tmp;) {
		foo=tmp;
		tmp=(p_captype)tmp->l.next;
		free(foo);
		foo=NULL;
	}

	/* Remove all capdevs */
	for (dev=capdevs; dev; dev=(p_capdev)dev->l.next) {
		if ( dev->state == CAP_STATE_NULL ) continue;

		if ( dev->thread ) pthread_cancel(dev->thread);
		dev->capture->end(dev->priv);
		dev->state = CAP_STATE_NULL;
	}
	for ( dev=capdevs; dev;) {
		bar=dev;
		dev=(p_capdev)dev->l.next;
		free(bar->interface);
		free(bar);
		bar=NULL;
	}
	
	CRITICAL_STOP(captlock);
}

p_captype capdev_find_type(char *name)
{
	p_captype tmp=NULL;
	if ( name == NULL ) return NULL;
	for (tmp=captype; tmp; tmp=(p_captype)tmp->l.next) {
		if ( strcmp(name, tmp->name) == 0 ) {
			return tmp;
		}
	}
	return NULL;
}


int capdev_add_type( char *name, proc_capdev_init i, 
       proc_capdev_end e, proc_capdev_poll p)
{
	p_captype type=NULL;
	
	if ( !name || !i || !e || !p ) return 0;
	
	if ( !(type=malloc(sizeof(captype_t))) ) return 0;

	type->name=name;
	type->init=i;
	type->end=e;
	type->poll=p;
	
	/* Add the captype */
	CRITICAL_START(captlock);
	captype=list_append(type, captype);
	CRITICAL_STOP(captlock);

	return 1;
}
