#include <conf.h>
#include <s_util.h>
#include <sensor.h>

conf_kw_t keywords[]={
	{"load_plugins", 	confh_load_plugins},
	{"capdev_add", 		confh_capdev_add},
	{"load_ruleset", 	confh_load_ruleset},
	{"firestorm_root",	confh_firestorm_root},
	{"chroot",		confh_chroot},
	{"effective_user", 	confh_euid},
	{"effective_group",	confh_egid},
	{"plugin_require", 	confh_require},
	{"global",		confh_global},
	{NULL, NULL}
};

kw_ret confh_require(int argc, char **argv)
{
	/* TODO */
	return KW_OK;
}

kw_ret confh_global(int argc, char **argv)
{
	char *k, *v;
	p_kv t;
	
	if ( argc != 3 ) return KW_ERROR;
	if ( !(t=malloc(sizeof(kv_t))) ) return KW_ERROR;
	if ( !(k=strdup(argv[1])) ) {
		free(t);
		return KW_ERROR;
	}
	if ( !(v=strdup(argv[2])) ) {
		free(t);
		free(k);
		return KW_ERROR;
	}
	
	t->key=k;
	t->val=v;
	globals.plugin_vars=list_append(t, globals.plugin_vars);
	
	return KW_OK;
}

kw_ret confh_euid(int argc, char **argv)
{
	static int done=0;

	if ( done ) return KW_MULTIPLE;
	if ( argc != 2 ) return KW_ERROR;

	globals.uid=atoi(argv[1]);
	return KW_OK;
}

kw_ret confh_egid(int argc, char **argv)
{
	static int done=0;

	if ( done ) return KW_MULTIPLE;
	if ( argc != 2 ) return KW_ERROR;

	globals.gid=atoi(argv[1]);
	return KW_OK;
}


kw_ret confh_chroot(int argc, char **argv)
{
	static int done=0;

	if ( done ) return KW_MULTIPLE;
	if ( argc != 2 ) return KW_ERROR;

	if ( !strcasecmp("yes", argv[1]) ) {
		globals.chroot=1;
	}else{
		globals.chroot=0;
	}
	
	done=1;
	return KW_OK;
}

kw_ret confh_firestorm_root(int argc, char **argv)
{
	static int done=0;

	if ( done ) return KW_MULTIPLE;
	if ( argc != 2 ) return KW_ERROR;

	if ( !(globals.root_dir=strdup(argv[1])) ) return KW_ERROR;
	
	done=1;
	return KW_OK;
}

kw_ret confh_load_ruleset(int argc, char **argv) {
	char *k, *v;
	p_kv t;
	
	if ( argc != 3 ) return KW_ERROR;
	if ( !(t=malloc(sizeof(kv_t))) ) return KW_ERROR;
	if ( !(k=strdup(argv[1])) ) {
		free(t);
		return KW_ERROR;
	}
	if ( !(v=strdup(argv[2])) ) {
		free(t);
		free(k);
		return KW_ERROR;
	}
	
	t->key=k;
	t->val=v;
	globals.rulesets=list_append(t, globals.rulesets);
	
	return KW_OK;
}

kw_ret confh_capdev_add(int argc, char **argv)
{
	p_c_capdev tmp;
	
	if ( argc!=3 && argc!=4 ) return KW_ERROR;
	if ( !(tmp=malloc(sizeof(c_capdev_t))) ) return KW_ERROR;
	
	tmp->type=strdup(argv[1]);
	tmp->source=strdup(argv[2]);
	tmp->filter = (argc==4) ? strdup(argv[3]) : NULL;
	
	if ( (tmp->type==NULL) ||
		(tmp->source==NULL) ||
		(argc==4 && tmp->filter==NULL) ) {
		if (tmp->type) free(tmp->type);
		if (tmp->source) free(tmp->source);
		if (tmp->filter) free(tmp->filter);
		return KW_ERROR;
	}

	globals.capdevs=list_append(tmp, globals.capdevs);
	
	return KW_OK;
}

kw_ret confh_load_plugins(int argc, char **argv)
{
	p_config_arg pdir;

	/* Only want 1 argument */
	if ( argc != 2 ) return KW_ERROR;
	
	if ( !(pdir=malloc(sizeof(config_arg_t))) ) return KW_ERROR;
	
	if ( !(pdir->arg=strdup(argv[1])) ){
		free(pdir);
		return KW_ERROR;
	}
	
	globals.plugin_dirs=list_append(pdir, globals.plugin_dirs);
	
	return KW_OK;
}

void conf_cleanup_args(p_config_arg p, int f)
{
	p_config_arg foo,bar;
	if (p==NULL) return;

	for(foo=p; foo; ){
		bar=foo;
		foo=(p_config_arg)bar->l.next;
		if ( f ) free(bar->arg);
		free(bar);
		bar=NULL;
	}
}

void conf_cleanup_kvs(int code, p_kv p)
{
	p_kv foo,bar;
	if (p==NULL) return;

	for(foo=p; foo; ){
		bar=foo;
		foo=(p_kv)bar->l.next;
		if ( bar->key ) free(bar->key);
		if ( bar->val ) free(bar->val);
		free(bar);
		bar=NULL;
	}
}

void conf_cleanup_caps(p_c_capdev p)
{
	p_c_capdev foo,bar;
	if (p==NULL) return;

	for(foo=p; foo; ){
		bar=foo;
		foo=(p_c_capdev)bar->l.next;
		if ( bar->type ) free(bar->type);
		if ( bar->source ) free(bar->source);
		if ( bar->filter ) free(bar->filter);
		free(bar);
		bar=NULL;
	}
}

proc_keyword_handler conf_get_handler(const char *kw)
{
	int i=0;
	proc_keyword_handler ret;
	
	if ( !kw ) return NULL;

	while ( (ret=keywords[i].proc) ) {
		if ( !strcasecmp(keywords[i].keyword, kw) ) {
			return ret;
		}
		i++;
	}
	
	return NULL;
}

int conf_parse_line(char *line)
{
	int ret;
	char *kw;
	char *tmp;
	p_config_arg args=NULL, foo;
	proc_keyword_handler kwh;
	
	int argc=1, i=1;
	char **argv;
	
	if ( !(kw=strtok(line, " \t\r\n")) ) return -1;

	if ( !(kwh=conf_get_handler(kw)) ) {
		print_out(PRINT_WARNING "Unrecognised keyword '%s'" CRLF, kw);
		return -2;
	}
	
	while ( (tmp=strtok(NULL, " \t\r\n")) ) {
		if ( !(foo=malloc(sizeof(config_arg_t))) ) {
			conf_cleanup_args(args, 0);
			return -1;
		}
		foo->arg=tmp;
		args=list_append(foo, args);
		argc++;
	}
	
	if ( !(argv=malloc(argc * sizeof(char *))) ) {
		conf_cleanup_args(args, 0);
		return -1;
	}
	
	argv[0]=kw;
	for ( foo=args; foo; foo=(p_config_arg)foo->l.next ) {
		argv[i]=foo->arg;
		i++;
	}
	
	ret = kwh(argc, argv);
	conf_cleanup_args(args, 0);
	free(argv);
	return ret;
}

int conf_load_config(char *fn)
{
	FILE *f;
	int line=0;
	char buf[1024], *p;
	
	if ( !fn ) return 0;
	
	if ( (f=fopen(fn, "r")) == NULL ) {
		print_out(PRINT_ERROR "Can't open config file %s" CRLF, fn);
		return 0;
	}
	
	for ( p=(char *)&buf; fgets(p, 4096, f); line++ ) {
		if ( !IS_CONFIG(buf) ) continue;

		switch(conf_parse_line(p)) {
			case KW_OK:
			case -2:
				break;
				
			case -1:
				print_out(PRINT_WARNING
					"Internal error in config file (%s) "
					"on line %i" CRLF, fn, line);
				return 0;

			case KW_ERROR:
				print_out(PRINT_WARNING
					"Parse error in config file (%s) "
					"on line %i" CRLF, fn, line);
				return 0;

			case KW_MULTIPLE:
				print_out(PRINT_WARNING
					"Problem in config file (%s) on line "
					"%i. This directive can only be "
					"used once." CRLF, fn, line);
				return 0;
		}
	}
	
	return 1;
}

int conf_go(void)
{
	p_config_arg tmp=NULL;
	p_c_capdev cd=NULL;
	p_kv kv=NULL;
	
	cleanup_add_handler(conf_cleanup_kvs, (void *)globals.plugin_vars);
	
	/* Chdir to firestorm_root first */
	if ( globals.root_dir ) {
		if ((chdir(globals.root_dir) != 0)) {
			print_out(PRINT_ERROR "Could't change to "
				"firstorm root dir (%s)" CRLF,
			      	globals.root_dir);
			return 0;
		}
	}

	/* Load plugins */
	for ( tmp=globals.plugin_dirs; tmp; tmp=(p_config_arg)tmp->l.next ) {
		if ( !loader_load_dir(tmp->arg) ) {
			print_out(PRINT_ERROR "EEK! Couldn't "
				"load plugin dir (%s)" CRLF, tmp->arg);
			return 0;
		}
	}conf_cleanup_args(globals.plugin_dirs, 1);
	globals.plugin_dirs=NULL;
	print_out(CRLF);
	
	/* Add and initialise all capdevs */
	for ( cd=globals.capdevs; cd; cd=(p_c_capdev)cd->l.next ) {
		capdev_add(cd->type, cd->source, cd->filter, capture_proc);
	}conf_cleanup_caps(globals.capdevs);
	globals.capdevs=NULL;
	capdev_init_all();
	
	/* Load rulesets */
	for (kv=globals.rulesets; kv; kv=(p_kv)kv->l.next ) {
		rules_loadfile(kv->key, kv->val);
	}conf_cleanup_kvs(0, globals.rulesets);
	globals.rulesets=NULL;
	
	/* Chroot if necessary */
	if ( globals.chroot ) {
		print_out("Chrooting to %s" CRLF, globals.root_dir);
		if ( chroot(".") == -1 ) {
			print_out(PRINT_ERROR "EEK! Couldn't "
				"chroot" CRLF);
			return 0;
		}
	}
	free(globals.root_dir);
	globals.root_dir=NULL;
	
	/* Drop UID and GID */
	if ( globals.gid ) {
		print_out("Dropping GID to %i" CRLF, globals.gid);
		if ( setgroups(1, &globals.gid) == -1 ||
			setgid(globals.gid) == -1) {
			print_out(PRINT_ERROR "Unexpected error "
				"while changing GID" CRLF);
			return 0;
		}
	}

	if ( globals.uid ) {
		print_out("Dropping UID to %i" CRLF, globals.uid);
		if ( setuid(globals.uid) == -1 ) {
			print_out(PRINT_ERROR "Unexpected error "
				"while changing UID" CRLF);
			return 0;
		}
	}
	
	return 1;
}
